/*
  averages_online.hpp

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _AVERAGES_ONLINE_HPP_
#define _AVERAGES_ONLINE_HPP_

template <typename T>
class generic_average_online {
public:
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
  typedef long double internal_value_type;
#else
  typedef T internal_value_type;
#endif
  generic_average_online() : value(0) {}
  virtual ~generic_average_online() {}
  virtual T average(const T v) = 0;
  const T get_value() const { return value; }
  virtual void reset() { value = internal_value_type(0); }
protected:
  internal_value_type value;
};

template <typename T>
class arithmetic_average_online : public generic_average_online<T> {
public:
  arithmetic_average_online() : count(0) {}
  T average(const T v)
  {
#if 0
    generic_average_online<T>::value = (generic_average_online<T>::value*count + v)/++count;
#else
    generic_average_online<T>::value += (v - generic_average_online<T>::value)/++count;
#endif
    return generic_average_online<T>::value; // restarts when 'count' will become zero again.
  }
  void reset() { generic_average_online<T>::reset(); count = 0; }
  unsigned long long get_count() const { return count; }
private:
  unsigned long long count;
};

template <typename T>
class geometric_average_online : public generic_average_online<T> {
public:
  geometric_average_online() : count(0), value(0) {}
  T average(const T v)
  {
#if 0
    generic_average_online<T>::value = count++ ? exp((log(generic_average_online<T>::value)*(count - 1) + log(v))/count) : v;
#else
    value += (log(v) - value)/++count;
    generic_average_online<T>::value = exp(value);
#endif
    return generic_average_online<T>::value; // restarts when 'count' will become zero again.
  }
  void reset() { generic_average_online<T>::reset(); count = 0; value = 0; }
  unsigned long long get_count() const { return count; }
private:
  unsigned long long count;
  typename generic_average_online<T>::internal_value_type value;
};

template <typename T>
class harmonic_average_online : public generic_average_online<T> {
public:
  harmonic_average_online() : count(0) {}
  T average(const T v)
  {
    generic_average_online<T>::value = count++ ? count/((count - 1)/generic_average_online<T>::value + T(1)/v) : v;
    return generic_average_online<T>::value; // restarts when 'count' will become zero again.
  }
  void reset() { generic_average_online<T>::reset(); count = 0; }
  unsigned long long get_count() const { return count; }
private:
  unsigned long long count;
};

template <typename T>
class generalized_average_online : public generic_average_online<T> {
public:
  generalized_average_online(const T p = 2) : count(0), power(p), value(0) {}
  T average(const T v)
  {
#if 0
    generic_average_online<T>::value = pow((pow(generic_average_online<T>::value, power)*count + pow(v, power))/++count, T(1)/power);
#else
    value += (pow(v, power) - value)/++count;
    generic_average_online<T>::value = pow(value, T(1)/power);
#endif
    return generic_average_online<T>::value; // restarts when 'count' will become zero again.
  }
  void reset() { generic_average_online<T>::reset(); count = 0; value = 0; }
  unsigned long long get_count() const { return count; }
private:
  unsigned long long count;
  const T power;
  typename generic_average_online<T>::internal_value_type value;
};

#if !(__cplusplus < 201103L) || defined(BOOST_FUNCTION_PROLOGUE_HPP)

#if !(__cplusplus < 201103L) && !defined(USE_BOOST_FUNCTION)
using std::function;
#else
using boost::function;
#endif

template <typename T>
class arbitrary_generalized_average_online : public generic_average_online<T> {
public:
  arbitrary_generalized_average_online(const function<T (T)> unary_op = log, const function<T (T)> inverse_unary_op = exp) : count(0), unary_operation(unary_op), inverse_unary_operation(inverse_unary_op), value(0) {}
  T average(const T v)
  {
#if 0
    generic_average_online<T>::value = count++ ? inverse_unary_operation((unary_operation(generic_average_online<T>::value)*(count - 1) + unary_operation(v))/count) : v;
#else
    value += (unary_operation(v) - value)/++count;
    generic_average_online<T>::value = inverse_unary_operation(value);
#endif
    return generic_average_online<T>::value; // restarts when 'count' will become zero again.
  }
  void reset() { generic_average_online<T>::reset(); count = 0; value = 0; }
  unsigned long long get_count() const { return count; }
private:
  unsigned long long count;
  const function<T (T)> unary_operation, inverse_unary_operation;
  typename generic_average_online<T>::internal_value_type value;
};

#endif

template <typename T>
class exponential_smoothed_average_online : public generic_average_online<T> {
public:
  exponential_smoothed_average_online(T factor = .1/*\in [0, 1]*/) : initial(true), smoothing_factor(factor) {}
  T average(const T v)
  {
    if (initial) {
      generic_average_online<T>::value = v;
      initial = false;
    }
    else
      generic_average_online<T>::value = generic_average_online<T>::value*(1 - smoothing_factor) + v*smoothing_factor;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; }
  bool is_initial() const { return initial; }
private:
  bool initial;
  T smoothing_factor;
};

template <typename T>
class moving_arithmetic_average_online : public generic_average_online<T> {
public:
  moving_arithmetic_average_online(const size_t n = 100) : size(n), initial(true), index(0)
  {
    storage = new T[size];
  }
  virtual ~moving_arithmetic_average_online()
  {
    delete [] storage;
  }
  T average(const T v)
  {
    if (initial) {
#if 0
      generic_average_online<T>::value = (generic_average_online<T>::value*index + v)/(index+1);
#else
      generic_average_online<T>::value += (v - generic_average_online<T>::value)/(index+1);
#endif
      storage[index++] = v;
      if (index == size) initial = false;
    }
    else {
      generic_average_online<T>::value -= storage[index]/size;
      generic_average_online<T>::value += (storage[index++] = v)/size;
    }
    if (index == size) index = 0;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; index = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const T *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  T *storage;
};

template <typename T>
class moving_geometric_average_online : public generic_average_online<T> {
public:
  moving_geometric_average_online(const size_t n = 100) : size(n), initial(true), index(0), value(0)
  {
    storage = new T[size];
  }
  virtual ~moving_geometric_average_online()
  {
    delete [] storage;
  }
  T average(const T v)
  {
    if (initial) {
#if 0
      generic_average_online<T>::value = index ? exp((log(generic_average_online<T>::value)*index + log(v))/(index+1)) : v;
#else
      value += (log(v) - value)/(index+1);
      generic_average_online<T>::value = exp(value);
#endif
      storage[index++] = v;
      if (index == size) initial = false;
    }
    else {
      value -= log(storage[index])/size;
      value += log((storage[index++] = v))/size;
      generic_average_online<T>::value = exp(value);
    }
    if (index == size) index = 0;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; index = 0; value = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const T *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  T *storage;
private:
  typename generic_average_online<T>::internal_value_type value;
};

template <typename T>
class moving_harmonic_average_online : public generic_average_online<T> {
public:
  moving_harmonic_average_online(const size_t n = 100) : size(n), initial(true), index(0)
  {
    storage = new T[size];
  }
  virtual ~moving_harmonic_average_online()
  {
    delete [] storage;
  }
  T average(const T v)
  {
    if (initial) {
      generic_average_online<T>::value = index ? (index+1)/(index/generic_average_online<T>::value + 1./v) : v;
      storage[index++] = v;
      if (index == size) initial = false;
    }
    else {
      generic_average_online<T>::value = 1./generic_average_online<T>::value;
      generic_average_online<T>::value -= (1./storage[index])/size;
      generic_average_online<T>::value += (1./(storage[index++] = v))/size;
      generic_average_online<T>::value = 1./generic_average_online<T>::value;
    }
    if (index == size) index = 0;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; index = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const T *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  T *storage;
};

template <typename T>
class moving_generalized_average_online : public generic_average_online<T> {
public:
  moving_generalized_average_online(const size_t n = 100, const T p = 2) : size(n), initial(true), index(0), power(p)
  {
    storage = new T[size];
  }
  virtual ~moving_generalized_average_online()
  {
    delete [] storage;
  }
  T average(const T v)
  {
    if (initial) {
#if 0
      generic_average_online<T>::value = pow((pow(generic_average_online<T>::value, power)*index + pow(v, power))/(index+1), 1./power);
#else
      value += (pow(v, power) - value)/(index+1);
      generic_average_online<T>::value = pow(value, 1./power);
#endif
      storage[index++] = v;
      if (index == size) initial = false;
    }
    else {
      value -= pow(storage[index], power)/size;
      value += pow((storage[index++] = v), power)/size;
      generic_average_online<T>::value = pow(value, 1./power);
    }
    if (index == size) index = 0;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; index = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const T *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  T *storage;
protected:
  const T power;
  typename generic_average_online<T>::internal_value_type value;
};

#if !(__cplusplus < 201103L) || defined(BOOST_FUNCTION_PROLOGUE_HPP)

#if !(__cplusplus < 201103L) && !defined(USE_BOOST_FUNCTION)
using std::function;
#else
using boost::function;
#endif

template <typename T>
class moving_arbitrary_generalized_average_online : public generic_average_online<T> {
public:
  moving_arbitrary_generalized_average_online(const size_t n = 100, const function<T (T)> unary_op = log, const function<T (T)> inverse_unary_op = exp) : size(n), initial(true), index(0), unary_operation(unary_op), inverse_unary_operation(inverse_unary_op), value(0)
  {
    storage = new T[size];
  }
  virtual ~moving_arbitrary_generalized_average_online()
  {
    delete [] storage;
  }
  T average(const T v)
  {
    if (initial) {
#if 0
      generic_average_online<T>::value = index ? inverse_unary_operation((unary_operation(generic_average_online<T>::value)*index + unary_operation(v))/(index+1)) : v;
#else
      value += (unary_operation(v) - value)/(index+1);
      generic_average_online<T>::value = inverse_unary_operation(value);
#endif
      storage[index++] = v;
      if (index == size) initial = false;
    }
    else {
      value -= unary_operation(storage[index])/size;
      value += unary_operation((storage[index++] = v))/size;
      generic_average_online<T>::value = inverse_unary_operation(value);
    }
    if (index == size) index = 0;
    return generic_average_online<T>::value;
  }
  void reset() { generic_average_online<T>::reset(); initial = true; index = 0; value = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const T *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  T *storage;
private:
  const function<T (T)> unary_operation, inverse_unary_operation;
  typename generic_average_online<T>::internal_value_type value;
};

#endif

#endif
