/*
  gsl_logarithmic_histogram_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <cmath>
#if !0 // with boost accumulators
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/density.hpp>
#include <numeric>

#include <boost/accumulators/statistics/positive_min.hpp>
#include <boost/accumulators/statistics/logarithmic_density.hpp>

namespace boost { namespace accumulators
{

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define internal_real_type long double
#else
#define internal_real_type double
#endif

template <typename T>
T histogram_mean(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
		 const size_t n_points)
{
  internal_real_type w(0), v(0);
  for (typename std::vector<std::pair<T, T> >::iterator it=histogram.begin()+1; it!=histogram.end()-1; ++it) {
    if (!(it->second > 0)) continue;
    const internal_real_type p((it->first + (it + 1)->first)/2);
    w += it->second*n_points;
    v += (p - v)*(it->second*n_points/w);
  }
  return v;
}
template <typename T>
T histogram_variance(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
		     const size_t n_points)
{
  internal_real_type mu(histogram_mean<T>(histogram, n_points)), w(0), v(0);
  for (typename std::vector<std::pair<T, T> >::iterator it=histogram.begin()+1; it!=histogram.end()-1; ++it) {
    if (!(it->second > 0)) continue;
    const internal_real_type p((it->first + (it + 1)->first)/2), delta(p - mu);
    w += it->second*n_points;
    v += (delta*delta - v)*(it->second*n_points/w);
  }
  return v;
}
template <typename T>
T histogram_standard_deviation(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
			       const size_t n_points)
{
  return sqrt(histogram_variance<T>(histogram, n_points));
}

template <typename T>
T histogram_logarithm_mean(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
			   const size_t n_points)
{
  internal_real_type w(0), v(0);
  for (typename std::vector<std::pair<T, T> >::iterator it=histogram.begin()+1; it!=histogram.end()-1; ++it) {
    if (!(it->second > 0)) continue;
    const internal_real_type p(exp((log(it->first) + log((it + 1)->first))/2));
    w += it->second*n_points;
    v += (p - v)*(it->second*n_points/w);
  }
  return v;
}
template <typename T>
T histogram_logarithm_variance(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
			       const size_t n_points)
{
  internal_real_type mu(histogram_logarithm_mean<T>(histogram, n_points)), w(0), v(0);
  for (typename std::vector<std::pair<T, T> >::iterator it=histogram.begin()+1; it!=histogram.end()-1; ++it) {
    if (!(it->second > 0)) continue;
    const internal_real_type p(exp((log(it->first) + log((it + 1)->first))/2)), delta(p - mu);
    w += it->second*n_points;
    v += (delta*delta - v)*(it->second*n_points/w);
  }
  return v;
}
template <typename T>
T histogram_logarithm_standard_deviation(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
					 const size_t n_points)
{
  return sqrt(histogram_logarithm_variance<T>(histogram, n_points));
}

template <typename T>
typename std::vector<std::pair<T, T> >::iterator histogram_min_element(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram)
{
#if __cplusplus < 201103L
  if (histogram.begin()+1 == histogram.end()-1) return histogram.begin()+1;
  typename std::vector<std::pair<T, T> >::iterator it(histogram.begin()+1), rv(it);
  while (++it != histogram.end()-1)
    if (it->second < rv->second)
      rv = it;
  return rv;
#else
  return min_element(histogram.begin()+1, histogram.end()-1,
		     [](const std::pair<T, T> &a, const typename std::pair<T, T> &b) ->
		     bool { return a.second < b.second; });
#endif
}
template <typename T>
ptrdiff_t histogram_min_index(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram)
{
  return std::distance(histogram.begin()+1, histogram_min_element<T>(histogram)-1);
}
template <typename T>
T histogram_min_value(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
		      const size_t n_points)
{
  return round(histogram_min_element<T>(histogram)->second*n_points);
}

template <typename T>
typename std::vector<std::pair<T, T> >::iterator histogram_max_element(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram)
{
#if __cplusplus < 201103L
  if (histogram.begin()+1 == histogram.end()-1) return histogram.begin()+1;
  typename std::vector<std::pair<T, T> >::iterator it(histogram.begin()+1), rv(it);
  while (++it != histogram.end()-1)
    if (!(it->second < rv->second))
      rv = it;
  return rv;
#else
  return max_element(histogram.begin()+1, histogram.end()-1,
		     [](const std::pair<T, T> &a, const std::pair<T, T> &b) ->
		     bool { return a.second < b.second; });
#endif
}
template <typename T>
ptrdiff_t histogram_max_index(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram)
{
  return std::distance(histogram.begin()+1, histogram_max_element<T>(histogram)-1);
}
template <typename T>
T histogram_max_value(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
		      const size_t n_points)
{
  return round(histogram_max_element<T>(histogram)->second*n_points);
}

template <typename T>
T histogram_summation(boost::iterator_range<typename std::vector<std::pair<T, T> >::iterator > &histogram,
		      const size_t n_points)
{
  return n_points - (round(histogram[0].second*n_points) + round(histogram[histogram.size()-1].second*n_points));
}

}}

#endif

#include <gsl/gsl_errno.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_histogram.h>

double gsl_stats_positive_min(const double data[], const size_t stride, const size_t n)
{
  size_t i = 0, rv = n;
  while (i != n) {
    if (0 < data[i*stride]) {
      rv = i++;
      break;
    }
    ++i;
  }
  while (i != n) {
    if (0 < data[i*stride] && data[i*stride] < data[rv*stride])
      rv = i;
    ++i;
  }
  return rv != n ? data[rv*stride] : 0;
}
double gsl_stats_negative_max(const double data[], const size_t stride, const size_t n)
{
  size_t i = 0, rv = n;
  while (i != n) {
    if (data[i*stride] < 0) {
      rv = i++;
      break;
    }
    ++i;
  }
  while (i != n) {
    if (data[i*stride] < 0 && data[rv*stride] < data[i*stride])
      rv = i;
    ++i;
  }
  return rv != n ? data[rv*stride] : 0;
}

int gsl_histogram_set_ranges_logarithm(gsl_histogram *h, double lb, double ub)
{
  size_t i;
#if 0
  const double log_lb = log(lb), log_ub = log(ub);

  for (i=0; i<h->n+1; i++)
    h->range[i] = exp((double)(h->n - i)/h->n*log_lb + (double)i/h->n*log_ub);
#else
  const long double log_lb = logl(lb), log_ub = logl(ub);

  for (i=0; i<h->n+1; i++)
    h->range[i] = expl((long double)(h->n - i)/h->n*log_lb + (long double)i/h->n*log_ub);
#endif
  for (i=0; i<h->n; i++)
    h->bin[i] = 0;
  return GSL_SUCCESS;
}

int gsl_histogram_set_ranges_logarithm_lt_ub(gsl_histogram *h, double lb, double ub)
{
  size_t i;
#if 0
  const double log_lb = log(lb), log_ub = log(ub), sz = log_ub - log_lb;

  for (i=0; i<h->n+1; i++)
    h->range[i] = lb*exp(sz*i/h->n);
#else
  const long double log_lb = logl(lb), log_ub = logl(ub), sz = log_ub - log_lb;

  for (i=0; i<h->n+1; i++)
    h->range[i] = lb*expl(sz*i/h->n);
#endif
  for (i=0; i<h->n; i++)
    h->bin[i] = 0;
  return GSL_SUCCESS;
}

double gsl_histogram_logarithm_mean(const gsl_histogram *h)
{
  const size_t n = h->n;
  size_t i;
  long double wmean = 0;
  long double W = 0;

  for (i = 0; i < n; i++) {
    double xi = exp((log(h->range[i + 1]) + log(h->range[i])) / 2);
    double wi = h->bin[i];

    if (wi > 0) {
      W += wi;
      wmean += (xi - wmean) * (wi / W);
    }
  }
  return wmean;
}

double gsl_histogram_logarithm_sigma(const gsl_histogram *h)
{
  const size_t n = h->n;
  size_t i;
  long double wvariance = 0 ;
  long double wmean = 0;
  long double W = 0;

  for (i = 0; i < n; i++) {
    double xi = exp((log(h->range[i + 1]) + log(h->range[i])) / 2);
    double wi = h->bin[i];

    if (wi > 0) {
      W += wi;
      wmean += (xi - wmean) * (wi / W);
    }
  }
  W = 0.0;
  for (i = 0; i < n; i++) {
    double xi = exp((log(h->range[i + 1]) + log(h->range[i])) / 2);
    double wi = h->bin[i];

    if (wi > 0) {
      const long double delta = (xi - wmean);
      W += wi ;
      wvariance += (delta * delta - wvariance) * (wi / W);
    }
  }
  {
    double sigma = sqrt(wvariance);
    return sigma;
  }
}

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sort.h>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#if __cplusplus < 201103L || defined(USE_BOOST_UTILITY)
#include <boost/utility.hpp>
#endif
#include "averages.hpp"
#include "statistics.hpp"
#include "histograms.hpp"
#include "histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

double distribution_type_value(const size_t t, gsl_rng *r)
{
  double rv = 0;

  switch (t) {
  case 0:	rv = gsl_rng_uniform(r);	break;
  case 1:	rv = gsl_ran_ugaussian(r);	break;
  case 2:	rv = gsl_ran_gaussian(r, 1.);	break;
  case 3:	rv = gsl_ran_ugaussian_tail(r, 1.5);	break;
  case 4:	rv = gsl_ran_gaussian_tail(r, 1.5, 1.);	break;
  case 5:	rv = gsl_ran_exponential(r, 1.);	break;
  case 6:	rv = gsl_ran_laplace(r, 1.);	break;
  case 7:	rv = gsl_ran_exppow(r, 1., 2.5);	break;
  case 8:	rv = gsl_ran_cauchy(r, 1.);	break;
  case 9:	rv = gsl_ran_rayleigh(r, 1.);	break;
  case 10:	rv = gsl_ran_rayleigh_tail(r, 1., 1.);	break;
  case 11:	rv = gsl_ran_landau(r);	break;
  case 12:	rv = gsl_ran_levy(r, 1., 1.);	break;
  case 13:	rv = gsl_ran_levy_skew(r, 1., 1., 1.);	break;
  case 14:	rv = gsl_ran_gamma(r, 1., 1.);	break;
  case 15:	rv = gsl_ran_flat(r, 0., 1.);	break;
  case 16:	rv = gsl_ran_chisq(r, 1.);	break;
  case 17:	rv = gsl_ran_lognormal(r, 0., 1.);	break;
  case 18:	rv = gsl_ran_fdist(r, 3., 2.);	break;
  case 19:	rv = gsl_ran_tdist(r, 1.);	break;
  case 20:	rv = gsl_ran_beta(r, 2., 2.);	break;
  case 21:	rv = gsl_ran_logistic(r, 1.);	break;
  case 22:	rv = gsl_ran_pareto(r, 1., 1.);	break;
  case 23:	rv = gsl_ran_weibull(r, 1., 1.);	break;
  case 24:	rv = gsl_ran_gumbel1(r, 1., 1.);	break;
  case 25:	rv = gsl_ran_gumbel2(r, 1., 1.);	break;
  case 26:	rv = gsl_ran_poisson(r, 2.5);	break;
  case 27:	rv = gsl_ran_bernoulli(r, .7) + 1;	break;
  case 28:	rv = gsl_ran_binomial(r, .5, 9);	break;
  case 29:	rv = gsl_ran_negative_binomial(r, .5, 3.5);	break;
  case 30:	rv = gsl_ran_pascal(r, .5, 3);	break;
  case 31:	rv = gsl_ran_geometric(r, .5);	break;
  case 32:	rv = gsl_ran_hypergeometric(r, 5, 20, 3);	break;
  default:	rv = gsl_ran_logarithmic(r, .7);	break;
  }
  return rv;
}
const char *distribution_names[] = {
  "uniform",
  "ugaussian",
  "gaussian",
  "ugaussian_tail",
  "gaussian_tail",
  "exponential",
  "laplace",
  "exppow",
  "cauchy",
  "rayleigh",
  "rayleigh_tail",
  "landau",
  "levy",
  "levy_skew",
  "gamma",
  "flat",
  "chisq",
  "lognormal",
  "fdist",
  "tdist",
  "beta",
  "logistic",
  "pareto",
  "weibull",
  "gumbel1",
  "gumbel2",
  "poisson",
  "bernoulli",
  "binomial",
  "negative_binomial",
  "pascal",
  "geometric",
  "hypergeometric",
  "logarithmic",
};

int main()
{
  const double maximum_permissible_error(
					 std::numeric_limits<double>::min()
					 );
#if !0 // with boost accumulators
  const double boost_maximum_permissible_error(
					       //std::numeric_limits<double>::min()
					       //1e-5 // histogram mean
					       //1e-4 // histogram sigma
					       //1e-3 // histogram mean
					       //1e-2 // histogram sigma
					       1e-0 // histogram mean
					       );
  const double continue_maximum_permissible_error(
						  //std::numeric_limits<double>::min()
						  1e-15 // histogram sigma
						  );
  const double online_maximum_permissible_error(
						//std::numeric_limits<double>::min()
						//1e-2 // histogram mean
						1e-0 // histogram min_value
						);
#endif
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram min_element",
    "histogram min_value",
    "histogram max_element",
    "histogram max_value",
    "histogram mean",
    "histogram sigma",
    "histogram sum",
  };
  const size_t n(1000000), n_bins(1000), n_prior_points(10000);
  double *a = (double *)malloc(sizeof(double)*n);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::vector<std::string> tabs_for_result_names(tabs_for_string_list(result_names, result_names + numberof(result_names), 1));

  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);

  histogram_online<double> histogram_online(n_bins, n_prior_points, true, true);
  
  for (size_t t=0; t<33; ++t) {
    std::vector<double> v;
    std::cout << distribution_names[t] << ":" << std::endl;
#if !0 // with boost accumulators
    boost::accumulators::accumulator_set<
      double,
      boost::accumulators::stats<
	boost::accumulators::tag::logarithmic_density
	>
      > accumulator_set(
			boost::accumulators::logarithmic_density_cache_size = n_prior_points, boost::accumulators::logarithmic_density_num_bins = n_bins
			);
#endif
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      a[i] = distribution_type_value(t, r);
#if !0 // with boost accumulators
      accumulator_set(a[i]);
#endif
      histogram_online.add(a[i]);
      v.push_back(a[i]);
    }

    double gsl_result_range[] = {
      gsl_stats_positive_min(a, 1, n),
      gsl_stats_max(a, 1, n),
    };
    gsl_histogram *gsl_hst = gsl_histogram_alloc(n_bins);
    gsl_histogram_set_ranges_logarithm(gsl_hst, gsl_result_range[0], gsl_result_range[1]);
    for (size_t i=0; i<n; ++i)
      gsl_histogram_increment(gsl_hst, a[i]);
    double gsl_results[] = {
      gsl_result_range[0],
      gsl_result_range[1],
      (double)gsl_histogram_min_bin(gsl_hst),
      gsl_histogram_min_val(gsl_hst),
      (double)gsl_histogram_max_bin(gsl_hst),
      gsl_histogram_max_val(gsl_hst),
      gsl_histogram_logarithm_mean(gsl_hst),
      gsl_histogram_logarithm_sigma(gsl_hst),
      gsl_histogram_sum(gsl_hst),
    };

    double result_range[] = {
      positive_min_value(v.begin(), v.end()),
      *max_element(v.begin(), v.end()),
    };
    std::vector<size_t> histogram_result(n_bins);
    size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), result_range[0], result_range[1], true));
    double results[] = {
      result_range[0],
      result_range[1],
      (double)std::distance(histogram_result.begin(), min_element(histogram_result.begin(), histogram_result.end())),
      (double)*min_element(histogram_result.begin(), histogram_result.end()),
      (double)std::distance(histogram_result.begin(), max_element(histogram_result.begin(), histogram_result.end())),
      (double)*max_element(histogram_result.begin(), histogram_result.end()),
      histogram_mean(histogram_result, result_range[0], result_range[1], true),
      histogram_standard_deviation(histogram_result, result_range[0], result_range[1], true),
#if !0
      (double)histogram_result_summation,
#else
      (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0)),
#endif
    };

    double continue_result_range[] = {
      positive_min_value(v.begin(), v.begin() + n_prior_points),
      *max_element(v.begin(), v.begin() + n_prior_points),
    };
    std::vector<size_t> histogram_continue_result(n_bins);
    size_t histogram_continue_result_summation(histogram(histogram_continue_result, v.begin(), v.begin() + n_prior_points, continue_result_range[0], continue_result_range[1], true));
    histogram_continue_result_summation += histogram(histogram_continue_result, v.begin() + n_prior_points, v.end(), continue_result_range[0], continue_result_range[1], true);
    double continue_results[] = {
      positive_min_value(v.begin(), v.end()),
      *max_element(v.begin(), v.end()),
      (double)std::distance(histogram_continue_result.begin(), min_element(histogram_continue_result.begin(), histogram_continue_result.end())),
      (double)*min_element(histogram_continue_result.begin(), histogram_continue_result.end()),
      (double)std::distance(histogram_continue_result.begin(), max_element(histogram_continue_result.begin(), histogram_continue_result.end())),
      (double)*max_element(histogram_continue_result.begin(), histogram_continue_result.end()),
      histogram_mean(histogram_continue_result, continue_result_range[0], continue_result_range[1], true),
      histogram_standard_deviation(histogram_continue_result, continue_result_range[0], continue_result_range[1], true),
#if !0
      (double)histogram_continue_result_summation,
#else
      (double)std::accumulate(histogram_continue_result.begin(), histogram_continue_result.end(), size_t(0)),
#endif
    };

    double online_results[] = {
      histogram_online.get_positive_minimum_value(),
      histogram_online.get_maximum_value(),
      (double)histogram_online.get_minimum_height_index(),
      histogram_online.get_minimum_height(),
      (double)histogram_online.get_maximum_height_index(),
      histogram_online.get_maximum_height(),
      histogram_online.get_heights_mean(),
      histogram_online.get_heights_standard_deviation(),
      (double)histogram_online.get_heights_summation(),
    };

#if !0 // with boost accumulators
    double boost_result_range[] = {
      boost::accumulators::positive_min(accumulator_set),
      boost::accumulators::max(accumulator_set),
    };
    boost::iterator_range<std::vector<std::pair<double, double> >::iterator >
      boost_histogram_result(boost::accumulators::logarithmic_density(accumulator_set));
    double boost_results[] = {
      boost_result_range[0],
      boost_result_range[1],
#if __cplusplus < 201103L
      (double)(boost::accumulators::histogram_min_index<double>(boost_histogram_result)+1),
      boost::accumulators::histogram_min_value<double>(boost_histogram_result, n),
      (double)(boost::accumulators::histogram_max_index<double>(boost_histogram_result)+1),
      boost::accumulators::histogram_max_value<double>(boost_histogram_result, n),
#else
      (double)std::distance(boost_histogram_result.begin()+1, min_element(boost_histogram_result.begin()+1, boost_histogram_result.end()-1,
									  [](const decltype(boost_histogram_result)::value_type &a, const decltype(boost_histogram_result)::value_type &b) ->
									  bool { return a.second < b.second; })),
      round(min_element(boost_histogram_result.begin()+1, boost_histogram_result.end()-1,
			[](const decltype(boost_histogram_result)::value_type &a, const decltype(boost_histogram_result)::value_type &b) ->
			bool { return a.second < b.second; })->second*n),
      (double)std::distance(boost_histogram_result.begin()+1, max_element(boost_histogram_result.begin()+1, boost_histogram_result.end()-1,
									  [](const decltype(boost_histogram_result)::value_type &a, const decltype(boost_histogram_result)::value_type &b) ->
									  bool { return a.second < b.second; })),
      round(max_element(boost_histogram_result.begin()+1, boost_histogram_result.end()-1,
			[](const decltype(boost_histogram_result)::value_type &a, const decltype(boost_histogram_result)::value_type &b) ->
			bool { return a.second < b.second; })->second*n),
#endif
      boost::accumulators::histogram_logarithm_mean<double>(boost_histogram_result, n),
      boost::accumulators::histogram_logarithm_standard_deviation<double>(boost_histogram_result, n),
#if __cplusplus < 201103L
      boost::accumulators::histogram_summation<double>(boost_histogram_result, n),
#else
      round(std::accumulate(boost_histogram_result.begin()+1, boost_histogram_result.end()-1, double(0),
			    [](const decltype(boost_histogram_result)::value_type::second_type &a, const decltype(boost_histogram_result)::value_type &b) ->
			    double { return a + b.second*n; })),
#endif
    };
#endif
    std::cout << "gsl_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(gsl_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;
    for (size_t b=0; b<n_bins; ++b) {
      if (!(gsl_hst->range[b] == histogram_range(b, n_bins, result_range[0], result_range[1], true).first &&
	    gsl_hst->bin[b] == histogram_result[b]))
	std::cout << "gsl_hst->bin[] - histogram_result[]:\t" << b << "\t" <<
	  gsl_hst->range[b] << "\t" << gsl_hst->bin[b] << "\t" <<
	  histogram_range(b, n_bins, result_range[0], result_range[1], true).first << "\t" << histogram_result[b] << std::endl;
      assert(gsl_hst->range[b] == histogram_range(b, n_bins, result_range[0], result_range[1], true).first);
      assert(gsl_hst->bin[b] == histogram_result[b]);
    }

#if !0 // with boost accumulators
    std::cout << "boost_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(boost_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << boost_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      if (!(j==2||j==4)) assert(error < boost_maximum_permissible_error);
    }
    std::cout << std::endl;

    std::cout << "boost_results[i] - continue_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(continue_results); ++j) {
      double error(sqrt(pow(boost_results[j] - continue_results[j], 2)) / (continue_results[j] ? fabs(continue_results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << boost_results[j] << "\t" << continue_results[j] << "\t" << error << std::endl;
      assert(error < continue_maximum_permissible_error);
    }
    std::cout << std::endl;
    for (size_t b=0; b<n_bins; ++b) {
      if (!(boost_histogram_result[b+1].first == histogram_lower_bound(b, n_bins, continue_result_range[0], continue_result_range[1], true, true) &&
	    abs(round(boost_histogram_result[b+1].second*n) - histogram_continue_result[b]) < 1))
	std::cout << "boost_histogram_result[] - histogram_continue_result[]:\t" << b << "\t" <<
	  boost_histogram_result[b+1].first << "\t" << round(boost_histogram_result[b+1].second*n) << "\t" <<
	  histogram_lower_bound(b, n_bins, continue_result_range[0], continue_result_range[1], true, true) << "\t" << histogram_continue_result[b] << std::endl;
      assert(boost_histogram_result[b+1].first == histogram_lower_bound(b, n_bins, continue_result_range[0], continue_result_range[1], true, true));
      assert(abs(round(boost_histogram_result[b+1].second*n) - histogram_continue_result[b]) < 1);
    }

    std::cout << "boost_results[i] - online_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(online_results); ++j) {
      double error(sqrt(pow(boost_results[j] - online_results[j], 2)) / (online_results[j] ? fabs(online_results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << boost_results[j] << "\t" << online_results[j] << "\t" << error << std::endl;
      assert(error < online_maximum_permissible_error);
    }
    std::cout << std::endl;
    for (size_t b=0; b<n_bins; ++b) {
      if (!(boost_histogram_result[b+1].first == histogram_online.get_lower_bound(b) &&
	    abs(round(boost_histogram_result[b+1].second*n) - histogram_online.get_height(b)) < 1))
	std::cout << "boost_histogram_result[] - histogram_online_result[]:\t" << b << "\t" <<
	  boost_histogram_result[b+1].first << "\t" << round(boost_histogram_result[b+1].second*n) << "\t" <<
	  histogram_online.get_lower_bound(b) << "\t" << histogram_online.get_height(b) << std::endl;
      assert(boost_histogram_result[b+1].first == histogram_online.get_lower_bound(b));
      assert(abs(round(boost_histogram_result[b+1].second*n) - histogram_online.get_height(b)) < 1);
    }
#endif

    histogram_online.reset();

    gsl_histogram_free(gsl_hst);
  }
  gsl_rng_free(r);
  return 0;
}
