/*
  gsl_multidimensional_histograms_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <cmath>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_histogram2d.h>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#include "multidimensional_histograms.hpp"
#include "container_numeric.hpp"
#include "multidimensional_histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

void bidimensional_distribution_type_value(const size_t t, gsl_rng *r, double *px, double *py)
{
  switch (t) {
  case 0:	*px = gsl_rng_uniform(r); *py = gsl_rng_uniform(r);	break;
  case 1:	gsl_ran_bivariate_gaussian(r, 1., 1., 1., px, py);	break;
  case 2:	gsl_ran_dir_2d(r, px, py);	break;
  default:	gsl_ran_dir_2d_trig_method(r, px, py);	break;
  }
}
const char *bidimensional_distribution_names[] = {
  "bidimensional uniform",
  "bivariate_gaussian",
  "dir_2d",
  "dir_2d_trig_method",
};

int main()
{
  const double maximum_permissible_error(
					 //std::numeric_limits<double>::min()
					 //1e-20 // histogram covariance
					 //1e-18 // histogram covariance
					 1e-13 // histogram mean
					 );
  const double continue_maximum_permissible_error(
						  //std::numeric_limits<double>::min()
						  1e-2 // histogram covariance
						  );
  const double online_maximum_permissible_error(
						std::numeric_limits<double>::min()
						//1e-2 // histogram covariance
						);
  const char *scalar_result_names[] = {
    "histogram min_value",
    "histogram max_value",
    "histogram sum",
    "histogram covariance",
  };
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram min_element",
    "histogram max_element",
    "histogram mean",
    "histogram sigma",
  };
  
  const size_t n_dimension(2), n(1000000), n_bins(1000), n_prior_points(100000);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::vector<std::string> tabs_for_scalar_result_names(tabs_for_string_list(scalar_result_names, scalar_result_names + numberof(scalar_result_names), 1));
  
  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);

  multidimensional_histogram_online<double, size_t, bool> histogram_online(n_bins, n_dimension, n_prior_points, false, true);

  for (size_t t=0; t<4; ++t) {
    std::vector<std::vector<double> > a(n);
    std::cout << bidimensional_distribution_names[t] << ":" << std::endl;
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      std::vector<double> v(2);
      bidimensional_distribution_type_value(t, r, &v[0], &v[1]);
      histogram_online.add(v);
      a[i] = v;
    }

    std::vector<double> gsl_result_range[] = {
      min_values(a.begin(), a.end()),
      max_values(a.begin(), a.end()),
    };
    gsl_histogram2d *gsl_hst = gsl_histogram2d_alloc(n_bins, n_bins);
    gsl_histogram2d_set_ranges_uniform(gsl_hst,
				       gsl_result_range[0][0], gsl_result_range[1][0],
				       gsl_result_range[0][1], gsl_result_range[1][1]);
    for (size_t i=0; i<n; ++i)
      gsl_histogram2d_increment(gsl_hst, a[i][0], a[i][1]);
    double gsl_scalar_results[] = {
      gsl_histogram2d_min_val(gsl_hst),
      gsl_histogram2d_max_val(gsl_hst),
      gsl_histogram2d_sum(gsl_hst),
      gsl_histogram2d_cov(gsl_hst),
    };
    std::vector<double> gsl_results[numberof(result_names)] = {
      gsl_result_range[0],
      gsl_result_range[1],
    };
    {
      size_t min_bin[2], max_bin[2];
      gsl_histogram2d_min_bin(gsl_hst, &min_bin[0], &min_bin[1]);
      gsl_histogram2d_max_bin(gsl_hst, &max_bin[0], &max_bin[1]);
      double more_results[][2] = {
	{ (double)min_bin[0],	(double)min_bin[1] },
	{ (double)max_bin[0],	(double)max_bin[1] },
	{ gsl_histogram2d_xmean(gsl_hst),	gsl_histogram2d_ymean(gsl_hst) },
	{ gsl_histogram2d_xsigma(gsl_hst),	gsl_histogram2d_ysigma(gsl_hst) },
      };
      for (size_t i=0; i<numberof(more_results); ++i)
	gsl_results[2+i] = std::vector<double>(more_results[i], more_results[i] + numberof(more_results[i]));
    }

    std::vector<double> result_range[] = {
      min_values(a.begin(), a.end()),
      max_values(a.begin(), a.end()),
    };
    std::map<std::vector<size_t>, size_t> histogram_result;
    size_t histogram_result_summation(histogram(histogram_result, a.begin(), a.end(), n_bins, result_range[0], result_range[1], false));
    double scalar_results[] = {
#if __cplusplus < 201103L
      (double)histogram_minimum_height(histogram_result),
      (double)histogram_maximum_height(histogram_result),
#else
      (double)min_element(histogram_result.begin(), histogram_result.end(),
			  [](const decltype(histogram_result)::value_type &a, const decltype(histogram_result)::value_type &b) ->
			  bool { return a.second < b.second; })->second,
      (double)max_element(histogram_result.begin(), histogram_result.end(),
			  [](const decltype(histogram_result)::value_type &a, const decltype(histogram_result)::value_type &b) ->
			  bool { return a.second < b.second; })->second,
#endif
#if __cplusplus < 201103L
      (double)histogram_result_summation,
#else
      (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0),
			      [](const decltype(histogram_result)::value_type::second_type &a, const decltype(histogram_result)::value_type &b) ->
			      decltype(histogram_result)::value_type::second_type { return a + b.second; }),
#endif
      histogram_covariance(histogram_result, n_bins, result_range[0], result_range[1], false),
    };
    std::vector<double> results[] = {
      result_range[0],
      result_range[1],
#if __cplusplus < 201103L
      container_cast<std::vector<double> >(histogram_minimum_height_index(histogram_result)),
      container_cast<std::vector<double> >(histogram_maximum_height_index(histogram_result)),
#else
      container_cast<std::vector<double> >(min_element(histogram_result.begin(), histogram_result.end(),
						       [](const decltype(histogram_result)::value_type &a, const decltype(histogram_result)::value_type &b) ->
						       bool { return a.second < b.second; })->first),
      container_cast<std::vector<double> >(max_element(histogram_result.begin(), histogram_result.end(),
						       [](const decltype(histogram_result)::value_type &a, const decltype(histogram_result)::value_type &b) ->
						       bool { return a.second < b.second; })->first),
#endif
      histogram_mean(histogram_result, n_bins, result_range[0], result_range[1], false),
      histogram_standard_deviation(histogram_result, n_bins, result_range[0], result_range[1], false),
    };

    std::vector<double> continue_result_range[] = {
      min_values(a.begin(), a.begin() + n_prior_points),
      max_values(a.begin(), a.begin() + n_prior_points),
    };
    std::map<std::vector<size_t>, size_t> histogram_continue_result;
    size_t histogram_continue_result_summation(histogram(histogram_continue_result, a.begin(), a.begin() + n_prior_points, n_bins, continue_result_range[0], continue_result_range[1], false));
    histogram_continue_result_summation += histogram(histogram_continue_result, a.begin() + n_prior_points, a.end(), n_bins, continue_result_range[0], continue_result_range[1], false);
    double scalar_continue_results[] = {
#if __cplusplus < 201103L
      (double)histogram_minimum_height(histogram_continue_result),
      (double)histogram_maximum_height(histogram_continue_result),
#else
      (double)min_element(histogram_continue_result.begin(), histogram_continue_result.end(),
			  [](const decltype(histogram_continue_result)::value_type &a, const decltype(histogram_continue_result)::value_type &b) ->
			  bool { return a.second < b.second; })->second,
      (double)max_element(histogram_continue_result.begin(), histogram_continue_result.end(),
			  [](const decltype(histogram_continue_result)::value_type &a, const decltype(histogram_continue_result)::value_type &b) ->
			  bool { return a.second < b.second; })->second,
#endif
#if __cplusplus < 201103L
      (double)histogram_continue_result_summation,
#else
      (double)std::accumulate(histogram_continue_result.begin(), histogram_continue_result.end(), size_t(0),
			      [](const decltype(histogram_continue_result)::value_type::second_type &a, const decltype(histogram_continue_result)::value_type &b) ->
			      decltype(histogram_continue_result)::value_type::second_type { return a + b.second; }),
#endif
      histogram_covariance(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
    };
    std::vector<double> continue_results[] = {
      min_values(a.begin(), a.end()),
      max_values(a.begin(), a.end()),
#if __cplusplus < 201103L
      container_cast<std::vector<double> >(histogram_minimum_height_index(histogram_continue_result)),
      container_cast<std::vector<double> >(histogram_maximum_height_index(histogram_continue_result)),
#else
      container_cast<std::vector<double> >(min_element(histogram_continue_result.begin(), histogram_continue_result.end(),
						       [](const decltype(histogram_continue_result)::value_type &a, const decltype(histogram_continue_result)::value_type &b) ->
						       bool { return a.second < b.second; })->first),
      container_cast<std::vector<double> >(max_element(histogram_continue_result.begin(), histogram_continue_result.end(),
						       [](const decltype(histogram_continue_result)::value_type &a, const decltype(histogram_continue_result)::value_type &b) ->
						       bool { return a.second < b.second; })->first),
#endif
      histogram_mean(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
      histogram_standard_deviation(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
    };

    double scalar_online_results[] = {
      histogram_online.get_minimum_height(),
      histogram_online.get_maximum_height(),
      (double)histogram_online.get_heights_summation(),
      histogram_online.get_heights_covariance(),
    };
    std::vector<double> online_results[] = {
      histogram_online.get_minimum_value(),
      histogram_online.get_maximum_value(),
      container_cast<std::vector<double> >(histogram_online.get_minimum_height_index()),
      container_cast<std::vector<double> >(histogram_online.get_maximum_height_index()),
      histogram_online.get_heights_mean(),
      histogram_online.get_heights_standard_deviation(),
    };

    std::cout << "gsl_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(scalar_result_names); ++j) {
      double error(sqrt(pow(gsl_scalar_results[j] - scalar_results[j], 2)) / (scalar_results[j] ? fabs(scalar_results[j]) : 1));
      std::cout << scalar_result_names[j] << ":" << tabs_for_scalar_result_names[j] << gsl_scalar_results[j] << "\t" << scalar_results[j] << "\t" << error << std::endl;
      if (!(j==0)) assert(error < maximum_permissible_error);
    }
    for (size_t j=0; j<numberof(result_names); ++j) {
      std::vector<double> error(sqrt(pow(gsl_results[j] - results[j], 2)));
      std::cout << result_names[j] << ":" << std::endl;
      for (size_t i=0; i<n_dimension; ++i)
	std::cout << "\t" << i << ":\t" << gsl_results[j][i] << "\t" << results[j][i] << "\t" << error[i] << std::endl;
      if (!(j==2)) assert(*max_element(error.begin(), error.end()) < maximum_permissible_error);
    }
    std::cout << std::endl;

    std::cout << "continue_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(scalar_result_names); ++j) {
      double error(sqrt(pow(scalar_continue_results[j] - scalar_results[j], 2)) / (scalar_results[j] ? fabs(scalar_results[j]) : 1));
      std::cout << scalar_result_names[j] << ":" << tabs_for_scalar_result_names[j] << scalar_continue_results[j] << "\t" << scalar_results[j] << "\t" << error << std::endl;
      if (!(j==1)) assert(error < continue_maximum_permissible_error);
    }
    for (size_t j=0; j<numberof(result_names); ++j) {
      std::vector<double> error(sqrt(pow(continue_results[j] - results[j], 2)));
      std::cout << result_names[j] << ":" << std::endl;
      for (size_t i=0; i<n_dimension; ++i)
	std::cout << "\t" << i << ":\t" << continue_results[j][i] << "\t" << results[j][i] << "\t" << error[i] << std::endl;
      if (!(j==3)) assert(*max_element(error.begin(), error.end()) < continue_maximum_permissible_error);
    }
    std::cout << std::endl;

    std::cout << "online_results[i] - continue_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(scalar_result_names); ++j) {
      double error(sqrt(pow(scalar_online_results[j] - scalar_continue_results[j], 2)) / (scalar_continue_results[j] ? fabs(scalar_continue_results[j]) : 1));
      std::cout << scalar_result_names[j] << ":" << tabs_for_scalar_result_names[j] << scalar_online_results[j] << "\t" << scalar_continue_results[j] << "\t" << error << std::endl;
      assert(error < online_maximum_permissible_error);
    }
    for (size_t j=0; j<numberof(result_names); ++j) {
      std::vector<double> error(sqrt(pow(online_results[j] - continue_results[j], 2)));
      std::cout << result_names[j] << ":" << std::endl;
      for (size_t i=0; i<n_dimension; ++i)
	std::cout << "\t" << i << ":\t" << online_results[j][i] << "\t" << continue_results[j][i] << "\t" << error[i] << std::endl;
      assert(*max_element(error.begin(), error.end()) < online_maximum_permissible_error);
    }
    std::cout << std::endl;

    histogram_online.reset();

    gsl_histogram2d_free(gsl_hst);
  }
  gsl_rng_free(r);
  return 0;
}
