/*
  gsl_running_quantiles_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cassert>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rstat.h>

void
gsl_rstat_quantile_reset(gsl_rstat_quantile_workspace *w)
{
  int i;
  double p = w->p;

  /* initialize positions n */
  for (i = 0; i < 5; ++i)
    w->npos[i] = i + 1;

  /* initialize n' */
  w->np[0] = 1.0;
  w->np[1] = 1.0 + 2.0 * p;
  w->np[2] = 1.0 + 4.0 * p;
  w->np[3] = 3.0 + 2.0 * p;
  w->np[4] = 5.0;

  /* initialize dn' */
  w->dnp[0] = 0.0;
  w->dnp[1] = 0.5 * p;
  w->dnp[2] = p;
  w->dnp[3] = 0.5 * (1.0 + p);
  w->dnp[4] = 1.0;

  w->n = 0;
} /* gsl_rstat_quantile_reset() */

#include <gsl/gsl_sort.h>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#if __cplusplus < 201103L || defined(USE_BOOST_UTILITY)
#include <boost/utility.hpp>
#endif
#include "quantiles.hpp"
#include "quantiles_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

double distribution_type_value(const size_t t, gsl_rng *r)
{
  double rv = 0;

  switch (t) {
  case 0:	rv = gsl_rng_uniform(r);	break;
  case 1:	rv = gsl_ran_ugaussian(r);	break;
  case 2:	rv = gsl_ran_gaussian(r, 1.);	break;
  case 3:	rv = gsl_ran_ugaussian_tail(r, 1.5);	break;
  case 4:	rv = gsl_ran_gaussian_tail(r, 1.5, 1.);	break;
  case 5:	rv = gsl_ran_exponential(r, 1.);	break;
  case 6:	rv = gsl_ran_laplace(r, 1.);	break;
  case 7:	rv = gsl_ran_exppow(r, 1., 2.5);	break;
  case 8:	rv = gsl_ran_cauchy(r, 1.);	break;
  case 9:	rv = gsl_ran_rayleigh(r, 1.);	break;
  case 10:	rv = gsl_ran_rayleigh_tail(r, 1., 1.);	break;
  case 11:	rv = gsl_ran_landau(r);	break;
  case 12:	rv = gsl_ran_levy(r, 1., 1.);	break;
  case 13:	rv = gsl_ran_levy_skew(r, 1., 1., 1.);	break;
  case 14:	rv = gsl_ran_gamma(r, 1., 1.);	break;
  case 15:	rv = gsl_ran_flat(r, 0., 1.);	break;
  case 16:	rv = gsl_ran_chisq(r, 1.);	break;
  case 17:	rv = gsl_ran_lognormal(r, 0., 1.);	break;
  case 18:	rv = gsl_ran_fdist(r, 3., 2.);	break;
  case 19:	rv = gsl_ran_tdist(r, 1.);	break;
  case 20:	rv = gsl_ran_beta(r, 2., 2.);	break;
  case 21:	rv = gsl_ran_logistic(r, 1.);	break;
  case 22:	rv = gsl_ran_pareto(r, 1., 1.);	break;
  case 23:	rv = gsl_ran_weibull(r, 1., 1.);	break;
  case 24:	rv = gsl_ran_gumbel1(r, 1., 1.);	break;
  case 25:	rv = gsl_ran_gumbel2(r, 1., 1.);	break;
  case 26:	rv = gsl_ran_poisson(r, 2.5);	break;
  case 27:	rv = gsl_ran_bernoulli(r, .7);	break;
  case 28:	rv = gsl_ran_binomial(r, .5, 9);	break;
  case 29:	rv = gsl_ran_negative_binomial(r, .5, 3.5);	break;
  case 30:	rv = gsl_ran_pascal(r, .5, 3);	break;
  case 31:	rv = gsl_ran_geometric(r, .5);	break;
  case 32:	rv = gsl_ran_hypergeometric(r, 5, 20, 3);	break;
  default:	rv = gsl_ran_logarithmic(r, .7);	break;
  }
  return rv;
}
const char *distribution_names[] = {
  "uniform",
  "ugaussian",
  "gaussian",
  "ugaussian_tail",
  "gaussian_tail",
  "exponential",
  "laplace",
  "exppow",
  "cauchy",
  "rayleigh",
  "rayleigh_tail",
  "landau",
  "levy",
  "levy_skew",
  "gamma",
  "flat",
  "chisq",
  "lognormal",
  "fdist",
  "tdist",
  "beta",
  "logistic",
  "pareto",
  "weibull",
  "gumbel1",
  "gumbel2",
  "poisson",
  "bernoulli",
  "binomial",
  "negative_binomial",
  "pascal",
  "geometric",
  "hypergeometric",
  "logarithmic",
};

int main()
{
  const double maximum_permissible_error(
					 std::numeric_limits<double>::min()
					 );
  const double gsl_maximum_permissible_error(
					     //std::numeric_limits<double>::min()
					     //1e-3 // uniform
					     //1e-2 // ugaussian
					     1e+0 // cauchy
					     );
#if 0 /* since 2016/01/29 */
#else  
  const double gsl_running_maximum_permissible_error(
						     //std::numeric_limits<double>::min()
						     //1e-4 // uniform
						     //1e-3 // ugaussian
						     1e+0 // cauchy
						     );
#endif
  const char *result_names[] = {
    "p=0.25 quantile",
    "p=0.5 quantile",
    "p=0.75 quantile",
    "0-quantile (least)",
    "2-quantile (median)",
    "1-quantile (greatest)",
  };
  const size_t n(100000);
  double *a = (double *)malloc(sizeof(double)*n);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::vector<std::string> tabs_for_result_names(tabs_for_string_list(result_names, result_names + numberof(result_names), 1));

  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);

  gsl_rstat_workspace *rstat_p = gsl_rstat_alloc();
  gsl_rstat_quantile_workspace *rstat_q[] = {
    gsl_rstat_quantile_alloc(0.25),
    gsl_rstat_quantile_alloc(0.5),
    gsl_rstat_quantile_alloc(0.75),
  };
#if 0 /* since 2016/01/29 */
  moving_quantiles_online<double> quantiles_online(n);
#else
  double probabilities[] = { .25, .5, .75, };
  heuristic_quantiles_online<double> quantiles_online(probabilities);
#endif

  std::vector<double> m(1000), s(m.size());
  moving_quantiles_online<double> moving_quantiles_online(m.size());

  for (size_t t=0; t<33; ++t) {
    std::vector<double> v;
    std::cout << distribution_names[t] << ":" << std::endl;
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      a[i] = distribution_type_value(t, r);
      v.push_back(a[i]);
      gsl_rstat_add(a[i], rstat_p);
      for (size_t k=0; k<numberof(rstat_q); ++k)
	gsl_rstat_quantile_add(a[i], rstat_q[k]);
      quantiles_online.add(a[i]);
    }
    gsl_sort(a, 1, n);
    //std::sort(&a[0], &a[n]);

    double gsl_results[] = {
      gsl_stats_quantile_from_sorted_data(a, 1, n, 0.25),
      gsl_stats_quantile_from_sorted_data(a, 1, n, 0.5),
      gsl_stats_quantile_from_sorted_data(a, 1, n, 0.75),
      a[0],
      gsl_stats_median_from_sorted_data(a, 1, n),
      a[n-1],
    };

    double gsl_running_results[] = {
      gsl_rstat_quantile_get(rstat_q[0]),
      gsl_rstat_quantile_get(rstat_q[1]),
      gsl_rstat_quantile_get(rstat_q[2]),
      gsl_rstat_min(rstat_p),
      gsl_rstat_median(rstat_p),
      gsl_rstat_max(rstat_p),
    };

    std::cout << "gsl_results[i] - gsl_running_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(gsl_results); ++j) {
      double error(sqrt(pow(gsl_results[j] - gsl_running_results[j], 2)));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << gsl_running_results[j] << "\t" << error << std::endl;
      assert(error < gsl_maximum_permissible_error);
    }
    std::cout << std::endl;

    std::sort(v.begin(), v.end());
    double results[] = {
#if !0
      quantile_in_sorted_container(v.begin(), v.end(), 0.25),
      quantile_in_sorted_container(v.begin(), v.end(), 0.5),
      quantile_in_sorted_container(v.begin(), v.end(), 0.75),
      //*v.begin(),
      least_in_sorted_container(v.begin(), v.end()),
      median_in_sorted_container(v.begin(), v.end()),
      //*(v.begin() + (std::distance(v.begin(), v.end()) - 1)),
      greatest_in_sorted_container(v.begin(), v.end()),
#else
      quantile_in_sorted_container(&a[0], &a[n], 0.25),
      quantile_in_sorted_container(&a[0], &a[n], 0.5),
      quantile_in_sorted_container(&a[0], &a[n], 0.75),
      //a[0],
      least_in_sorted_container(&a[0], &a[n]),
      median_in_sorted_container(&a[0], &a[n]),
      //a[n-1],
      greatest_in_sorted_container(&a[0], &a[n]),
#endif
    };
    std::cout << "gsl_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(gsl_results[j] - results[j], 2)));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;

    double online_results[] = {
      quantiles_online.get_quantile(0.25),
      quantiles_online.get_quantile(0.5),
      quantiles_online.get_quantile(0.75),
      quantiles_online.get_minimum_value(),
      quantiles_online.get_median(),
      quantiles_online.get_maximum_value(),
    };
#if 0 /* since 2016/01/29 */
    std::cout << "results[i] - online_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(results[j] - online_results[j], 2)));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << results[j] << "\t" << online_results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;
#else
    std::cout << "gsl_running_results[i] - online_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(gsl_running_results); ++j) {
      double error(sqrt(pow(gsl_running_results[j] - online_results[j], 2)));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_running_results[j] << "\t" << online_results[j] << "\t" << error << std::endl;
      assert(error < gsl_running_maximum_permissible_error);
    }
    std::cout << std::endl;
#endif

    gsl_rstat_quantile_reset(rstat_p->median_workspace_p);
    gsl_rstat_reset(rstat_p);
#if 0
    for (size_t k=0; k<numberof(rstat_q); ++k)
      gsl_rstat_quantile_free(rstat_q[k]);
    rstat_q[0] = gsl_rstat_quantile_alloc(0.25);
    rstat_q[1] = gsl_rstat_quantile_alloc(0.5);
    rstat_q[2] = gsl_rstat_quantile_alloc(0.75);
#else
    for (size_t k=0; k<numberof(rstat_q); ++k)
      gsl_rstat_quantile_reset(rstat_q[k]);
#endif
    quantiles_online.reset();

    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<m.size()*10; ++i) {
      double x(distribution_type_value(t, r));
      m[i % m.size()] = x;
      moving_quantiles_online.add(x);
      if (i < m.size()) continue;
      copy(m.begin(), m.end(), s.begin());
      sort(s.begin(), s.end());
      double moving_results[] = {
	quantile_in_sorted_container(s.begin(), s.end(), 0.25),
	quantile_in_sorted_container(s.begin(), s.end(), 0.5),
	quantile_in_sorted_container(s.begin(), s.end(), 0.75),
	//*s.begin(),
	least_in_sorted_container(s.begin(), s.end()),
	median_in_sorted_container(s.begin(), s.end()),
	//*(s.begin() + (std::distance(s.begin(), s.end()) - 1)),
	greatest_in_sorted_container(s.begin(), s.end()),
      };
      double moving_online_results[] = {
	moving_quantiles_online.get_quantile(0.25),
	moving_quantiles_online.get_quantile(0.5),
	moving_quantiles_online.get_quantile(0.75),
	moving_quantiles_online.get_minimum_value(),
	moving_quantiles_online.get_median(),
	moving_quantiles_online.get_maximum_value(),
      };

      if (i+1 == m.size()*10) std::cout << "moving_results[i] - moving_online_results[i]:" << std::endl;
      for (size_t j=0; j<numberof(moving_results); ++j) {
	double error(sqrt(pow(moving_results[j] - moving_online_results[j], 2)));
	if (i+1 == m.size()*10) {
	  std::cout << "moving " << result_names[j] << ":" << tabs_for_result_names[j] << moving_results[j] << "\t" << moving_online_results[j] << "\t" << error << std::endl;
	  if (j+1 == numberof(moving_results)) std::cout << std::endl;
	}
        assert(error < maximum_permissible_error);
      }
    }
    moving_quantiles_online.reset();
  }

  gsl_rng_free(r);
  return 0;
}
