/*
  multidimensional_histograms.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#if __cplusplus < 201103L
#include <boost/utility.hpp>
#endif
#include "statistics.hpp"
#include "multidimensional_histograms.hpp"
#include "multidimensional_histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

//#define USE_LOGARITHMIC_HISTOGRAM
//#define USE_LIST_FOR_MOVING_STORAGE

int main()
{
  const double maximum_permissible_error(std::numeric_limits<double>::min());
  const double online_maximum_permissible_error(1e-2);
  const char *scalar_result_names[] = {
    "histogram min_value",
    "histogram max_value",
    "histogram total",
    "histogram sum",
    "histogram covariance",
  };
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram lower bound",
    "histogram upper bound",
    "histogram min_element",
    "histogram max_element",
    "histogram mean",
    "histogram sigma",
  };
  std::vector<std::vector<double> > v(10000);
  const size_t n_dimension(2), n_bins(v.size()/1000), n_prior_points(v.size()/2);
#if 0 /* WITHOUT_MULTIDIMENSIONAL_HISTOGRAM_ONLINE */
#else
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  multidimensional_histogram_online<double, size_t, bool> histogram_online(n_bins, n_dimension, n_prior_points);
#else
  multidimensional_histogram_online<double, size_t, bool> histogram_online(n_bins, n_dimension, n_prior_points, true);
#endif
#endif
  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    v[i].resize(n_dimension);
    for (size_t j=0; j<n_dimension; ++j) {
      double x((double)random()/((1L<<31)-1));
      v[i][j] = x;
    }
#if 0 /* WITHOUT_MULTIDIMENSIONAL_HISTOGRAM_ONLINE */
#else
    histogram_online.add(v[i]);
#endif
  }
  std::vector<double> result_range[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    min_values(v.begin(), v.end()),
#else
    positive_min_values(v.begin(), v.end()),
#endif
    max_values(v.begin(), v.end()),
  };
  std::map<std::vector<size_t>, size_t> histogram_result;
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), n_bins, result_range[0], result_range[1], false));
#else
  size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), n_bins, result_range[0], result_range[1], true));
#endif
  double scalar_results[] = {
    (double)histogram_minimum_height(histogram_result),
    (double)histogram_maximum_height(histogram_result),
    (double)std::distance(v.begin(), v.end()),
#if __cplusplus < 201103L
    (double)histogram_result_summation,
#else
    (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0),
			    [](const decltype(histogram_result)::value_type::second_type &a, const decltype(histogram_result)::value_type &b) ->
			    decltype(histogram_result)::value_type::second_type { return a + b.second; }),
#endif
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_covariance(histogram_result, n_bins, result_range[0], result_range[1], false),
#else
    histogram_covariance(histogram_result, n_bins, result_range[0], result_range[1], true),
#endif
  };
  std::vector<double> results[] = {
    result_range[0],
    result_range[1],
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, result_range[0], result_range[1], false),
    histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, result_range[0], result_range[1], false),
#else
    histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, result_range[0], result_range[1], true),
    histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, result_range[0], result_range[1], true),
#endif
    container_cast<std::vector<double> >(histogram_minimum_height_index(histogram_result)),
    container_cast<std::vector<double> >(histogram_maximum_height_index(histogram_result)),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_mean(histogram_result, n_bins, result_range[0], result_range[1], false),
    histogram_standard_deviation(histogram_result, n_bins, result_range[0], result_range[1], false),
#else
    histogram_mean(histogram_result, n_bins, result_range[0], result_range[1], true),
    histogram_standard_deviation(histogram_result, n_bins, result_range[0], result_range[1], true),
#endif
  };
#if 0 /* WITHOUT_MULTIDIMENSIONAL_HISTOGRAM_ONLINE */
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  std::vector<double> continue_result_range[] = { min_values(v.begin(), v.begin() + n_prior_points), max_values(v.begin(), v.begin() + n_prior_points), };
#else
  std::vector<double> continue_result_range[] = { positive_min_values(v.begin(), v.begin() + n_prior_points), max_values(v.begin(), v.begin() + n_prior_points), };
#endif
  std::map<std::vector<size_t>, size_t> histogram_continue_result;
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  size_t histogram_continue_result_summation(histogram(histogram_continue_result, v.begin(), v.begin() + n_prior_points, n_bins, continue_result_range[0], continue_result_range[1], false));
  histogram_continue_result_summation += histogram(histogram_continue_result, v.begin() + n_prior_points, v.end(), n_bins, continue_result_range[0], continue_result_range[1], false);
#else
  size_t histogram_continue_result_summation(histogram(histogram_continue_result, v.begin(), v.begin() + n_prior_points, n_bins, continue_result_range[0], continue_result_range[1], true));
  histogram_continue_result_summation += histogram(histogram_continue_result, v.begin() + n_prior_points, v.end(), n_bins, continue_result_range[0], continue_result_range[1], true);
#endif
  double continue_scalar_results[] = {
    (double)histogram_minimum_height(histogram_continue_result),
    (double)histogram_maximum_height(histogram_continue_result),
    (double)std::distance(v.begin(), v.end()),
#if __cplusplus < 201103L
    (double)histogram_continue_result_summation,
#else
    (double)std::accumulate(histogram_continue_result.begin(), histogram_continue_result.end(), size_t(0),
			    [](const decltype(histogram_continue_result)::value_type::second_type &a, const decltype(histogram_continue_result)::value_type &b) ->
			    decltype(histogram_continue_result)::value_type::second_type { return a + b.second; }),
#endif
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_covariance(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
#else
    histogram_covariance(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], true),
#endif
  };
  std::vector<double> continue_results[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    min_values(v.begin(), v.end()),
#else
    positive_min_values(v.begin(), v.end()),
#endif
    max_values(v.begin(), v.end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, continue_result_range[0], continue_result_range[1], false),
    histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, continue_result_range[0], continue_result_range[1], false),
#else
    histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, continue_result_range[0], continue_result_range[1], true),
    histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, continue_result_range[0], continue_result_range[1], true),
#endif
    container_cast<std::vector<double> >(histogram_minimum_height_index(histogram_continue_result)),
    container_cast<std::vector<double> >(histogram_maximum_height_index(histogram_continue_result)),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_mean(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
    histogram_standard_deviation(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], false),
#else
    histogram_mean(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], true),
    histogram_standard_deviation(histogram_continue_result, n_bins, continue_result_range[0], continue_result_range[1], true),
#endif
  };
  for (size_t j=0; j<numberof(continue_scalar_results); ++j) {
    double error(sqrt(pow(continue_scalar_results[j] - scalar_results[j], 2)) / (scalar_results[j] ? fabs(scalar_results[j]) : 1));
    std::cout << scalar_result_names[j] << ":\t" << continue_scalar_results[j] << "\t" << scalar_results[j] << "\t" << error << std::endl;
    if (!(j==0||j==1||j==4)) assert(error < online_maximum_permissible_error);
  }
  for (size_t j=0; j<numberof(continue_results); ++j) {
    std::vector<double> error(sqrt(pow(continue_results[j] - results[j], 2)));
    std::cout << result_names[j] << ":" << std::endl;
    for (size_t i=0; i<n_dimension; ++i)
      std::cout << "\t" << i << ":\t" << continue_results[j][i] << "\t" << results[j][i] << "\t" << error[i] << std::endl;
    if (!(j==4||j==5)) assert(*max_element(error.begin(), error.end()) < online_maximum_permissible_error);
  }
#else
  double online_scalar_results[] = {
    (double)histogram_online.get_minimum_height(),
    (double)histogram_online.get_maximum_height(),
    (double)histogram_online.get_count(),
    (double)histogram_online.get_heights_summation(),
    histogram_online.get_heights_covariance(),
  };
  std::vector<double> online_results[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_online.get_minimum_value(),
#else
    histogram_online.get_positive_minimum_value(),
#endif
    histogram_online.get_maximum_value(),
    histogram_online.get_lower_bound(std::vector<size_t>(n_dimension, 0)),
    histogram_online.get_upper_bound(std::vector<size_t>(n_dimension, n_bins-1)),
    container_cast<std::vector<double> >(histogram_online.get_minimum_height_index()),
    container_cast<std::vector<double> >(histogram_online.get_maximum_height_index()),
#if !0
    histogram_online.get_heights_mean(),
    histogram_online.get_heights_standard_deviation(),
#else
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_mean(histogram_online.get_heights(), n_bins, histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1), false),
    histogram_standard_deviation(histogram_online.get_heights(), n_bins, histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1), false),
#else
    histogram_mean(histogram_online.get_heights(), n_bins, histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1), true),
    histogram_standard_deviation(histogram_online.get_heights(), n_bins, histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1), true),
#endif
#endif
  };
  for (size_t j=0; j<numberof(online_scalar_results); ++j) {
    double error(sqrt(pow(online_scalar_results[j] - scalar_results[j], 2)) / (scalar_results[j] ? fabs(scalar_results[j]) : 1));
    std::cout << scalar_result_names[j] << ":\t" << online_scalar_results[j] << "\t" << scalar_results[j] << "\t" << error << std::endl;
    if (!(j==0||j==1||j==4)) assert(error < online_maximum_permissible_error);
  }
  for (size_t j=0; j<numberof(online_results); ++j) {
    std::vector<double> error(sqrt(pow(online_results[j] - results[j], 2)));
    std::cout << result_names[j] << ":" << std::endl;
    for (size_t i=0; i<n_dimension; ++i)
      std::cout << "\t" << i << ":\t" << online_results[j][i] << "\t" << results[j][i] << "\t" << error[i] << std::endl;
    if (!(j==4||j==5)) assert(*max_element(error.begin(), error.end()) < online_maximum_permissible_error);
  }
#endif

#if !defined(USE_LIST_FOR_MOVING_STORAGE)
  std::vector<std::vector<double> > m(v.size());
#else
  std::list<std::vector<double> > m;
#endif
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  moving_multidimensional_histogram_online<double> moving_histogram_online(n_bins, n_dimension, v.size(), false);
#else
  moving_multidimensional_histogram_online<double> moving_histogram_online(n_bins, n_dimension, v.size(), true);
#endif
  for (size_t i=0; i<v.size()*2; ++i) {
    std::vector<double> a(n_dimension);
    for (size_t j=0; j<n_dimension; ++j) {
      double x((double)random()/((1L<<31)-1));
      a[j] = x;
    }
#if !defined(USE_LIST_FOR_MOVING_STORAGE)
    m[i % v.size()] = a;
#else
    if (!(i < v.size())) m.pop_front();
    m.push_back(a);
#endif
    moving_histogram_online.add(a);
    if (i < v.size()) continue;
    std::vector<double> moving_result_range[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      min_values(m.begin(), m.end()),
#else
      positive_min_values(m.begin(), m.end()),
#endif
      max_values(m.begin(), m.end()),
    };
    std::map<std::vector<size_t>, size_t> moving_histogram_result;
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    size_t moving_histogram_result_summation(histogram(moving_histogram_result, m.begin(), m.end(), n_bins, moving_result_range[0], moving_result_range[1], false));
#else
    size_t moving_histogram_result_summation(histogram(moving_histogram_result, m.begin(), m.end(), n_bins, moving_result_range[0], moving_result_range[1], true));
#endif
    double moving_scalar_results[] = {
      (double)histogram_minimum_height(moving_histogram_result),
      (double)histogram_maximum_height(moving_histogram_result),
      (double)std::distance(m.begin(), m.end()),
#if __cplusplus < 201103L
      (double)moving_histogram_result_summation,
#else
      (double)std::accumulate(moving_histogram_result.begin(), moving_histogram_result.end(), size_t(0),
			      [](const decltype(moving_histogram_result)::value_type::second_type &a, const decltype(moving_histogram_result)::value_type &b) ->
			      decltype(moving_histogram_result)::value_type::second_type { return a + b.second; }),
#endif
    };
    std::vector<double> moving_results[] = {
      moving_result_range[0],
      moving_result_range[1],
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, moving_result_range[0], moving_result_range[1], false),
      histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, moving_result_range[0], moving_result_range[1], false),
#else
      histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, moving_result_range[0], moving_result_range[1], true),
      histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, moving_result_range[0], moving_result_range[1], true),
#endif
      container_cast<std::vector<double> >(histogram_minimum_height_index(moving_histogram_result)),
      container_cast<std::vector<double> >(histogram_maximum_height_index(moving_histogram_result)),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      histogram_mean(moving_histogram_result, n_bins, moving_result_range[0], moving_result_range[1], false),
      histogram_standard_deviation(moving_histogram_result, n_bins, moving_result_range[0], moving_result_range[1], false),
#else
      histogram_mean(moving_histogram_result, n_bins, moving_result_range[0], moving_result_range[1], true),
      histogram_standard_deviation(moving_histogram_result, n_bins, moving_result_range[0], moving_result_range[1], true),
#endif
    };
    double online_moving_scalar_results[] = {
      (double)moving_histogram_online.get_minimum_height(),
      (double)moving_histogram_online.get_maximum_height(),
      (double)moving_histogram_online.get_count(),
#if __cplusplus < 201103L
      (double)moving_histogram_online.get_heights_summation(),
#else
      (double)std::accumulate(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end(), size_t(0),
			      [](const std::map<std::vector<size_t>, size_t>::value_type::second_type &a, const std::map<std::vector<size_t>, size_t>::value_type &b) ->
			      std::map<std::vector<size_t>, size_t>::value_type::second_type { return a + b.second; }),
#endif
    };
    std::vector<double> online_moving_results[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      moving_histogram_online.get_minimum_value(),
#else
      moving_histogram_online.get_positive_minimum_value(),
#endif
      moving_histogram_online.get_maximum_value(),
      moving_histogram_online.get_lower_bound(std::vector<size_t>(n_dimension, 0)),
      moving_histogram_online.get_upper_bound(std::vector<size_t>(n_dimension, n_bins-1)),
      container_cast<std::vector<double> >(moving_histogram_online.get_minimum_height_index()),
      container_cast<std::vector<double> >(moving_histogram_online.get_maximum_height_index()),
      moving_histogram_online.get_heights_mean(),
      moving_histogram_online.get_heights_standard_deviation(),
    };
    for (size_t j=0; j<numberof(online_moving_scalar_results); ++j) {
      double error(sqrt(pow(online_moving_scalar_results[j] - moving_scalar_results[j], 2)) / (moving_scalar_results[j] ? fabs(moving_scalar_results[j]) : 1));
      if (i+1 == v.size()*2 || !(error < maximum_permissible_error))
	std::cout << "moving " << scalar_result_names[j] << ":\t" << online_moving_scalar_results[j] << "\t" << moving_scalar_results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    for (size_t j=0; j<numberof(online_moving_results); ++j) {
      std::vector<double> error(sqrt(pow(online_moving_results[j] - moving_results[j], 2)));
      if (i+1 == v.size()*2 || !(*max_element(error.begin(), error.end()) < maximum_permissible_error)) {
        std::cout << "moving " << result_names[j] << ":" << std::endl;
	for (size_t i=0; i<n_dimension; ++i)
	  std::cout << "\t" << i << ":\t" << online_moving_results[j][i] << "\t" << moving_results[j][i] << "\t" << error[i] << std::endl;
      }
      assert(*max_element(error.begin(), error.end()) < maximum_permissible_error);
    }
  }
  return 0;
}
