/*
  pseudo_histograms.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#if __cplusplus < 201103L || defined(USE_BOOST_UTILITY)
#include <boost/utility.hpp>
#endif
#include "statistics.hpp"
#include "histograms.hpp"
#include "histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

//#define USE_LOGARITHMIC_HISTOGRAM

double modified_Bernoulli_map(double x, double b1 = 2, double b2 = 2)
{
  static const double epsilon(std::numeric_limits<double>::min());
  if (x <= .5)
    x += pow(2., b1-1)*(1 - 2*epsilon)*pow(x, b1) + epsilon;
  else
    x -= pow(2., b2-1)*(1 - 2*epsilon)*pow(1-x, b2) - epsilon;
  return x;
}

int main()
{
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  bool logarithmic(false);
#else
  bool logarithmic(true);
#endif
  const double maximum_permissible_error(1e-3);
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram lower bound",
    "histogram upper bound",
    "histogram min_element",
    "histogram min_value",
    "histogram max_element",
    "histogram max_value",
    "histogram mean",
    "histogram sigma",
    "histogram total",
    "histogram sum",
  };
  std::vector<double> v(10000000);
  const size_t n_bins(v.size()/10), n_prior_points(v.size());
  histogram_online<double> pseudo_dense_histogram_online(n_bins, n_prior_points, logarithmic);

  double x(.1);
  for (size_t i=0; i<v.size(); ++i) {
    x = modified_Bernoulli_map(x);
    v[i] = x;
  }
  double result_range[] = {
    !logarithmic ? *min_element(v.begin(), v.end()) : positive_min_value(v.begin(), v.end()),
    *max_element(v.begin(), v.end()),
  };
  std::vector<size_t> histogram_result(n_bins);
  size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), result_range[0], result_range[1], logarithmic));
  double results[] = {
    result_range[0],
    result_range[1],
    histogram_lower_bound(0, n_bins, result_range[0], result_range[1], logarithmic),
    histogram_upper_bound(n_bins-1, n_bins, result_range[0], result_range[1], logarithmic),
    (double)std::distance(histogram_result.begin(), min_element(histogram_result.begin(), histogram_result.end())),
    (double)*min_element(histogram_result.begin(), histogram_result.end()),
    (double)std::distance(histogram_result.begin(), max_element(histogram_result.begin(), histogram_result.end())),
    (double)*max_element(histogram_result.begin(), histogram_result.end()),
    histogram_mean(histogram_result, result_range[0], result_range[1], logarithmic),
    histogram_standard_deviation(histogram_result, result_range[0], result_range[1], logarithmic),
    (double)std::distance(v.begin(), v.end()),
#if !0
    (double)histogram_result_summation,
#else
    (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0)),
#endif
  };

#if !0 /* worse implement */
  std::vector<double> pd_histogram(histogram_cumulative_distribution<double>(histogram_result));
#else
  std::pair<std::vector<double>, std::vector<double> > pd_histogram(histogram_probability_cumulative_distribution<double>(histogram_result));
#endif

  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    double x((double)random()/((1L<<31)-1));
#if !0 /* worse implement */
    pseudo_dense_histogram_online.add(histogram_cumulative_distribution_pseudo(pd_histogram, x,
									       histogram_result, result_range[0], result_range[1], logarithmic));
#else
    pseudo_dense_histogram_online.add(histogram_probability_cumulative_distribution_pseudo(pd_histogram, x,
											   histogram_result, result_range[0], result_range[1], logarithmic));
#endif
  }
  
  double pd_results[] = {
    !logarithmic ? pseudo_dense_histogram_online.get_minimum_value() : pseudo_dense_histogram_online.get_positive_minimum_value(),
    pseudo_dense_histogram_online.get_maximum_value(),
    pseudo_dense_histogram_online.get_lower_bound(0),
    pseudo_dense_histogram_online.get_upper_bound(n_bins-1),
#if !0
    (double)pseudo_dense_histogram_online.get_minimum_height_index(),
    (double)pseudo_dense_histogram_online.get_minimum_height(),
    (double)pseudo_dense_histogram_online.get_maximum_height_index(),
    (double)pseudo_dense_histogram_online.get_maximum_height(),
    pseudo_dense_histogram_online.get_heights_mean(),
    pseudo_dense_histogram_online.get_heights_standard_deviation(),
    (double)pseudo_dense_histogram_online.get_count(),
    (double)pseudo_dense_histogram_online.get_heights_summation(),
#else
    (double)std::distance(pseudo_dense_histogram_online.get_heights().begin(), min_element(pseudo_dense_histogram_online.get_heights().begin(), pseudo_dense_histogram_online.get_heights().end())),
    (double)*min_element(pseudo_dense_histogram_online.get_heights().begin(), pseudo_dense_histogram_online.get_heights().end()),
    (double)std::distance(pseudo_dense_histogram_online.get_heights().begin(), max_element(pseudo_dense_histogram_online.get_heights().begin(), pseudo_dense_histogram_online.get_heights().end())),
    (double)*max_element(pseudo_dense_histogram_online.get_heights().begin(), pseudo_dense_histogram_online.get_heights().end()),
    histogram_mean(pseudo_dense_histogram_online.get_heights(), pseudo_dense_histogram_online.get_lower_bound(0), pseudo_dense_histogram_online.get_upper_bound(n_bins-1), logarithmic),
    histogram_standard_deviation(pseudo_dense_histogram_online.get_heights(), pseudo_dense_histogram_online.get_lower_bound(0), pseudo_dense_histogram_online.get_upper_bound(n_bins-1), logarithmic),
    (double)pseudo_dense_histogram_online.get_count(),
    (double)std::accumulate(pseudo_dense_histogram_online.get_heights().begin(), pseudo_dense_histogram_online.get_heights().end(), size_t(0)),
#endif
  };
  
  for (size_t j=0; j<numberof(pd_results); ++j) {
    double error(sqrt(pow(pd_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
    std::cout << result_names[j] << ":\t" << pd_results[j] << "\t" << results[j] << "\t" << error << std::endl;
    if (!(j==4||j==5||j==6||j==7)) assert(error < maximum_permissible_error);
  }

  return 0;
}
